@@ -2,6 +2,7 @@
 
 import os
 import sys
+import socket
 from optparse import OptionParser
 from wsgiref.simple_server import make_server, WSGIServer, WSGIRequestHandler
 
@@ -60,6 +61,10 @@ parser.add_option('--cache-force', default=False, dest='cache_force', action='st
     help='Force regeneration of tiles while in caching mode (defaults to False)'
     )
 
+parser.add_option('--processes', default=1, dest='num_processes', type='int',
+    help='If werkzeug is installed, number of rendering processes to allow'
+    )
+    
 def run(process):
     try:
         process.serve_forever()
@@ -75,6 +80,19 @@ def strip_opts(options):
             params[k] = v
     return params
 
+def print_url(options):
+    if not application.debug:
+        sys.stderr.write('TileLite debug mode is *off*...\n')
+    sys.stderr.write("Listening on %s:%s...\n" % (options.host,options.port))
+    sys.stderr.write("To access locally view: http://localhost:%s\n" % options.port)
+    remote = "To access remotely view: http://%s" % socket.getfqdn()
+    if not options.port == 80:
+        remote += ":%s" % options.port
+    remote += "\nor view: http://%s" % socket.gethostbyname(socket.gethostname())
+    if not options.port == 80:
+        remote += ":%s" % options.port
+    sys.stderr.write('%s\n' % remote)
+
 if __name__ == '__main__':
     (options, args) = parser.parse_args()
         
@@ -88,11 +106,11 @@ if __name__ == '__main__':
         if not os.path.exists(mapfile):
             sys.exit('Could not locate mapfile.')
     
-    print "Using mapfile: '%s'" % os.path.abspath(mapfile)
+    print "[TileLite Debug] --> Using mapfile: '%s'" % os.path.abspath(mapfile)
         
     if options.config:
         if not os.path.isfile(options.config):
-            sys.exit('That does not appear to be a config file')
+            sys.exit('That does not appear to be a valid config file')
         else:
             CONFIG = options.config
 
@@ -103,16 +121,15 @@ if __name__ == '__main__':
             CONFIG = None
     
     if CONFIG:
-        print "Using config file: '%s'" % os.path.abspath(CONFIG)        
+        sys.stderr.write("[TileLite Debug] --> Using config file: '%s'" % os.path.abspath(CONFIG))      
 
-    
     if options.cache_path and not options.caching:
         options.caching = True
 
     if options.cache_force and not options.caching:
         options.caching = True
 
-        #parser.error("Caching must be turned on with '--caching' flag for liteserv.py to accept '--cache-path' option")
+    #parser.error("Caching must be turned on with '--caching' flag for liteserv.py to accept '--cache-path' option")
     #http_setup = options.host, options.port
     #httpd = simple_server.WSGIServer(http_setup, WSGIRequestHandler)
     #httpd.set_app(application)
@@ -120,12 +137,27 @@ if __name__ == '__main__':
     from tilelite import Server
     application = Server(mapfile, CONFIG)
     application.absorb_options(strip_opts(options.__dict__))
-    
-    httpd = make_server(options.host, options.port, application)
-    print "Listening on %s:%s...\n" % (options.host,options.port)
-    print "If you are running locally view in your browser: 'http://localhost:%s'" % options.port
-    print "Otherwise go to: 'http://yourserver.com:%s'" % options.port
-    if not application.debug:
-        print 'TileLite debug mode is *off*...'
-    
-    run(httpd)
\ No newline at end of file
+                
+    try:
+        from werkzeug import run_simple
+        print_url(options)
+        run_simple(options.host, options.port, application, threaded=False, processes=options.num_processes)
+    except:
+        if options.num_processes > 1:
+            sys.exit('The werkzeug python server must be installed to run multi-process\n')
+        sys.stderr.write('Note: werkzeug is not installed so falling back to built-in python wsgiref server.\n')
+        sys.stderr.write('Install werkzeug from http://werkzeug.pocoo.org/\n\n')
+        
+        from wsgiref import simple_server
+        # below code was for testing multi-threaded rendering
+        # which only works if we copy a map object per thread
+        # so avoid this and only run multiprocess...
+        #from SocketServer import ThreadingMixIn
+        #class myServer(ThreadingMixIn, simple_server.WSGIServer):
+        #    pass 
+        #httpd = myServer(('',options.port), simple_server.WSGIRequestHandler,)
+        #httpd.set_app(application)
+        httpd = make_server(options.host, options.port, application)        
+        print_url(options)
+        run(httpd)
+        
